#ifndef __AMSCLIL_FBOPS_H__
#define __AMSCLIL_FBOPS_H__

///////////////////////////////
//AMS C Large Integer Library//
//Fixed Buffer Functions     //
///////////////////////////////

//Fixed buffer unsigned large integer operations
//
//These are needed, because preallocating the working memory
//will be much faster than dynamically allocating within iterative 
//operations

//For the convenience of the caller, these can be wrapped in 
//operations acting on resizeable buffers. For speed, these can
//be set up and called directly, with allocation/deallocation taking
//place at the start and end of the wrapping routine.

//Uint32 Operators

//   01234567890123456789012345678901 - MISRA 31 character identifier limit

//Single "digit" operations - these are sometimes reimplimented directly for speed
void amsclil1_ui32_add(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *car);
void amsclil1_ui32_sub(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *car);
void amsclil1_ui32_mult(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *car);
void amsclil1_ui32_div(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *rem);

//comparison:
//0: a and b are equal, 1: a is greater, 2: b is greater
int amsclil1_fb32_cmp(uint32_t* a, uint32_t* b, long blen);

int amsclil1_fb32_iszero(uint32_t* a, long blen);

void amsclil1_fb32_setzero(uint32_t *a, long blen);

//most significant index
long amsclil1_fb32_msi(uint32_t*a, long blen);

//large integer display functions
void amsclil1_fb32_liprint(uint32_t* a, long blen); //print hex representation
void amsclil1_fb16_liprint(uint16_t* a, long blen); //print hex representation
void amsclil1_fb32_libinprint(uint32_t* a, long blen); //print binary representation

//bit shift operations
//lshift: <<: *2^shift
//wrk[blen]
void amsclil1_fb32_shiftl(uint32_t* a, long shift, uint32_t* wrk, long blen);
void amsclil1_fb32_shiftr(uint32_t* a, long shift, uint32_t* wrk, long blen);

//Fixed Buffer Operators
void amsclil1_fb32_add(uint32_t *a, uint32_t *b, uint32_t *c, long blen);
void amsclil1_fb32_sub(uint32_t *a, uint32_t *b, uint32_t *c, long blen);

//schoolbook multiplication
// karatsuba multiplication may be needed for large enough numbers
void amsclil1_fb32_mult(uint32_t *a, uint32_t *b, uint32_t *c, uint32_t *wrk, long blen);

//Division using algorithm [HAC 14.20]
int amsclil1_fb32_div(uint32_t *a, uint32_t *b, uint32_t *res, uint32_t *rem, uint32_t* wrk, long blen);

//Implement!
void amsclil1_fb32_decimal(uint32_t *a, uint8_t *dd, uint32_t* wrk, int blen);

//Modular exponentiation [now using HAC14.42 Barrett Reduction]
void amsclil1_fb32_modpower(uint32_t *a, uint32_t *pw, uint32_t *md, uint32_t *res, uint32_t *wrk, uint32_t *stor, int blen);

//non-cryptographic random integer for testing
void amsclil1_fb32_ncrandint(uint32_t *a, long blen);

//Euler greatest common divisor algorithm
void amsclil1_fb32_eulergcd(uint32_t *a, uint32_t *b, uint32_t *gcd, uint32_t *wrk, long blen);

//Least Common Multiple
void amsclil1_fb32_lcm(uint32_t *a, uint32_t *b, uint32_t *lcm, uint32_t *wrk, long blen);

//Multiplicative Inverse Algorithm
int amsclil1_fb32_multinv(uint32_t *z, uint32_t *dom, uint32_t *zinv, uint32_t *wrk, int blen);


//Euler totient function
        // for prime numbers it's just n-1

//Carmichels Totient Function

////////////////////////////////////////
//Miscellaneous and Internal Functions//
////////////////////////////////////////

//16 bit fixed buffer operations

//used in HAC14.20 division algorithm
void amsclil1_fb16_mult(uint16_t *a, uint16_t *b, uint16_t *c, long blen);

//[HAC 14.42]: Barrett reduction of modular arithmetic
//Precomputation of mu divisor
void amsclil1_fb32_barrettmu(uint32_t *m, uint32_t *mu, uint32_t *wrk, long K, long blen);

//Barrett Reduced Modulus
//y = mod(x,m)
void amsclil1_fb32_barrettmod(uint32_t *x, uint32_t *m, uint32_t *y, 
        uint32_t *mu, uint32_t *wrk, long K, long Kx, long blen);

//Classical Fast Modular Exponentiation
void amsclil1_fb32_modpower_cls(uint32_t *a, uint32_t *pw, uint32_t *md, uint32_t *res, uint32_t *wrk, uint32_t *stor, int blen);
//Fast Modular Exponentiation with Barrett Reduction
void amsclil1_fb32_modpower_bar(uint32_t *a, uint32_t *pw, uint32_t *md, uint32_t *res, uint32_t *wrk, uint32_t *stor, int blen);

//Shifted Operations

//bit-shifted reads and writes
//don't move the array in memory, move the read/write
//perhaps it will save time within the division operation
int amsclil1_fb32_readshift(uint32_t* arr, long ind, long shift, uint32_t* val, long blen);
int amsclil1_fb32_writeshift(uint32_t* arr, long ind, long shift, uint32_t* val, long blen);

//Shifted Large Integer Display Function
void amsclil1_fb32_shiftliprint(uint32_t* a, long shift, long blen); //print hex representation
//Shifted integer, most significant index
long amsclil1_fb32_shiftmsi(uint32_t *a, long shift, long blen);
//Shifted comparison
int amsclil1_fb32_shiftcmp(uint32_t *a, uint32_t *b, long bshift, long blen);
//Shifted subtraction
void amsclil1_fb32_shiftsub(uint32_t *a, uint32_t *b, long bshift, uint32_t *c, long blen);



//internal digit shift subtraction
//this is supposed to be something like c = a - b*base^diff
void amsclil1_fb16_intlssub(uint16_t *a, uint16_t *b, long bshft, uint16_t *c, long blen);

//internal digit shift comparison
//this is supposed to be a <=> b*base^diff
int amsclil1_fb16_intlscmp(uint16_t *a, uint16_t *b, long bshft, long blen);

void amsclil1_fb16_intlmultsing(uint16_t *a, uint16_t b, long blen);


#endif