#include <amsclil1/amsclil1.h>

//precompute mu = b^(2*k)/m for Barrett Reduction algorithm [HAC 14.42]
//m[blen] - the modulus
//mu[2*(blen+1)] - precomputed factor for Barrett Reduction
//wrk[14*blen+19] - division working memory
//K - most significant index of m + 1 (at most blen)
//blen - length of m buffer
void amsclil1_fb32_barrettmu(uint32_t *m, uint32_t *mu, uint32_t *wrk, long K, long blen)
{
    long I;
    //intl buffer lengths 2*(blen+1)
    //division working length 5*(2*(blen+1)+1): 10*blen+15
    //total work length: 10*blen+15+4*(blen+1) 14*blen+19

    const size_t divlen = 10*blen+15;
    const size_t intblen = 2*(blen+1);
    uint32_t *b2 = &(wrk[divlen+0*intblen]);
    uint32_t *m2 = &(wrk[divlen+1*intblen]);
    uint32_t *mu2 = &(wrk[divlen+2*intblen]);
    uint32_t *rm2 = &(wrk[divlen+3*intblen]);
    
    for(I=0;I<blen;I++)
    {
        b2[I] = 0;
        m2[I] = m[I];
        mu2[I] = 0;
    }
    for(I=blen;I<2*(blen+1);I++)
    {
        b2[I] = 0;
        m2[I] = 0;
        mu2[I] = 0;
    }
    
    b2[2*K] = 1;

    //amsclil1_fb32_div(b2,m2,mu2,rm2,wrk,2*K+1);
    amsclil1_fb32_div(b2,m2,mu2,rm2,wrk,2*(blen+1));

    //if(amsclil1_fb32_iszero(mu2,2*(blen+1)))
    //{
    //    printf("debug:mu="); amsclil1_fb32_liprint(mu2,2*(blen+1)); printf("\n");
    //    printf("debug:b2="); amsclil1_fb32_liprint(b2,2*(blen+1)); printf("\n");
    //    printf("debug:m2="); amsclil1_fb32_liprint(m2,2*(blen+1)); printf("\n");
    //}
    
    //printf("debug:mu="); amsclil1_fb32_liprint(mu2,2*K+2); printf("\n");

    for(I=0;I<2*(blen+1);I++)
    {
        mu[I] = mu2[I];
    }

    return;
}

//compute y = mod(x,m) using Barrett Reduction Algorithm [HAC 14.42]
//x[blen] - x has (according to the algorithm) an msi of 2*k-1
//          here, supply only k-1 digits to keep buffers even
//m[blen] - m has msi of (K-1) which must be <- blen-2
//mu[2*(blen+1)] - precompute mu=(1<<(2*32*blen) / m)
//y[blen] - output
//wrk[33*(blen+1)] 
//K - provide most significant index of m +1
//Kx - provide most significant index of x +1
//blen - length of buffers
void amsclil1_fb32_barrettmod(uint32_t *x, uint32_t *m, uint32_t *y, 
        uint32_t *mu, uint32_t *wrk, long K, long Kx, long blen)
{
    long I;

    const size_t divlen = 5*(blen+1); //length of working buffer for division algorithm
    //offset for beginning of mod buffer
    const size_t intblen = 4*(blen+1);
    //wrk size: 5*blen+5 + 7*(4*blen+4): 33*(blen+1)

    int cmp;

    uint32_t* bswap = NULL;
    uint32_t* q1 = &(wrk[divlen+0*intblen]);
    uint32_t* q2 = &(wrk[divlen+1*intblen]);
    uint32_t* q3 = &(wrk[divlen+2*intblen]);
    uint32_t* r1 = &(wrk[divlen+3*intblen]);
    uint32_t* r2 = &(wrk[divlen+4*intblen]);
    uint32_t* r3 = &(wrk[divlen+5*intblen]);
    uint32_t* mi = &(wrk[divlen+6*intblen]);

    long dbg;

    //handle mod 1 and mod 0 cases fast
    if(K==1)
    {
        if(m[0]==0 || m[0]==1)
        {
            for(I=0;I<blen;I++)
            {
                y[I] = 0;
            }

            return;
        }
    }

    for(I=0;I<intblen;I++)
    {
        q1[I] = 0;
        q2[I] = 0; //mult takes care of that
        q3[I] = 0;
        r1[I] = 0;
        r2[I] = 0;
        r3[I] = 0;
        mi[I] = 0;
    }

    for(I=0;I<blen;I++) //copy m to larger internal buffer
    {
        mi[I] = m[I];
    }

    //1. q1 = x / b^(K-1)
    for(I=Kx-1;I>=(K-1);I--)
    //for(I=blen-1;I>=(K-1);I--)
    {
        q1[I-(K-1)] = x[I]; 
    }

    //2. q2 = q1*mu
    amsclil1_fb32_mult(mu,q1,q2,wrk,2*(blen+1));
    
    //3. q3 = q2/ b^(K+1)
    for(I=intblen-1;I>=K+1;I--)
    {
        q3[I-(K+1)] = q2[I];
    }

    //4. r1 = x mod b^(k+1)
    for(I=0;I<(K+1) && I<Kx;I++)
    {
        r1[I] = x[I];
    }

    //5. r2 = (q3 * m) mod b^(K+1)
    amsclil1_fb32_mult(q3,mi,r2,wrk,2*(blen+1));
    for(I=K+1;I<intblen;I++)
    {
        r2[I] = 0;
    }

    //printf("debug: r1="); amsclil1_fb32_liprint(r1,2*(blen+1)); printf("\n");
    //printf("debug: r2="); amsclil1_fb32_liprint(r2,2*(blen+1)); printf("\n");
    //printf("debug: md="); amsclil1_fb32_liprint(mi,2*(blen+1)); printf("\n");


    //6. if(r1<r2) r3 = r3 + b^(K+1)
    cmp = amsclil1_fb32_cmp(r1,r2,2*(blen+1));
    if(cmp==2)
    {
        r3[K+1] = 1;
    }
    //r3 = r3-r2+r1
    amsclil1_fb32_add(r3,r1,q1,2*(blen+1));
    amsclil1_fb32_sub(q1,r2,r3,2*(blen+1));

    //amsclil1_fb32_sub(r3,r2,q1,2*(blen+1));
    //amsclil1_fb32_add(q1,r1,r3,2*(blen+1));

    //printf("debug: r3="); amsclil1_fb32_liprint(r3,2*(blen+1)); printf("\n");

    //while r3>=m r3 = r3-m
    cmp = amsclil1_fb32_cmp(r3,mi,2*(blen+1));
    
    //dbg = 0;
    while(cmp!=2)
    {
        amsclil1_fb32_sub(r3,mi,q2,2*(blen+1));
        bswap = r3; r3 = q2; q2 = bswap;
        cmp = amsclil1_fb32_cmp(r3,mi,2*(blen+1));
        //dbg++;
    }
    
    for(I=0;I<blen;I++)
    {
        y[I] = r3[I];
    }

    //if(dbg>2)
    //{
    //    printf("debug: %ld iters, should be at most 2\n",dbg);
    //    printf("debug: x="); amsclil1_fb32_liprint(x,blen); printf("\n");
    //    printf("debug: m="); amsclil1_fb32_liprint(m,blen); printf("\n");
    //    printf("debug: y="); amsclil1_fb32_liprint(y,blen); printf("\n");
    //    printf("debug: mu="); amsclil1_fb32_liprint(mu,2*(blen+1)); printf("\n");
    //    printf("debug: r1="); amsclil1_fb32_liprint(r1,2*(blen+1)); printf("\n");
    //    printf("debug: r2="); amsclil1_fb32_liprint(r2,2*(blen+1)); printf("\n");
    //}

    return;
}


//Fast Modular Exponentiation using Barrett Reduction
//calculates mod(a^pw, md)
//a[blen] - base buffer
//pw[blen] - power
//md[blen] - modulus
//wrk[66*(blen+1)]
//stor[28*(blen+1)]
//blen - base buffer length
void amsclil1_fb32_modpower_bar(uint32_t *a, uint32_t *pw, uint32_t *md, uint32_t *res, uint32_t *wrk, uint32_t *stor, int blen)
{
    long I;
    long msipw, msia;
    uint32_t bit;
    long smaj;
    int smin;
    long Kmd,Kacc,Kbm;

    const size_t intblen = 4*(blen+1);

    uint32_t* bswap = NULL;
    uint32_t* bm0 = &(stor[0]);
    uint32_t* bm1 = &(stor[1*intblen]);
    uint32_t* ra = &(stor[2*intblen]);
    uint32_t* rb = &(stor[3*intblen]); //unused?
    uint32_t* acc = &(stor[4*intblen]);
    uint32_t* mdb = &(stor[5*intblen]);
    uint32_t* mu = &(stor[6*intblen]);

    msipw = amsclil1_fb32_msi(pw,blen);
    msia = amsclil1_fb32_msi(a,blen);

    Kmd = amsclil1_fb32_msi(md,blen)+1;

    for(I=0;I<blen;I++)
    {
        mdb[I] = md[I]; //init mdb to md
        acc[I] = 0;
        bm0[I] = a[I]; //init bm0 to a
        mu[I] = 0;
    }
    for(I=blen;I<intblen;I++)
    {
        mdb[I] = 0;
        acc[I] = 0;
        bm0[I] = 0;
        mu[I] = 0;
    }
    acc[0] = 1; //init acc to 1

    //precompute mu
    amsclil1_fb32_barrettmu(mdb, mu, wrk, Kmd, blen);
    //printf("debug: mu="); amsclil1_fb32_liprint(mu,2*blen); printf("\n");

    for(I=0;I<32*(msipw+1);I++)
    {
        smaj = I/32;
        smin = I%32;
        bit = (pw[smaj] & (1<<smin))>>smin;

        if(bit!=0)
        {
            amsclil1_fb32_mult(acc,bm0,ra,wrk,blen);
            bswap = acc; acc = ra; ra = bswap; //swap(acc,ra)
            Kacc = amsclil1_fb32_msi(acc,2*blen)+1;
            //printf("debug: acc="); amsclil1_fb32_liprint(acc,2*blen); printf("\n");
            //acc - [2*blen] buffer now
            //so everything fed to barrettmod needs to be sized for 2*blen?
            amsclil1_fb32_barrettmod(acc,mdb,ra,mu,wrk,Kmd,Kacc,2*blen);
            bswap = acc; acc = ra; ra = bswap; //swap(acc,ra)
            //printf("debug: acc="); amsclil1_fb32_liprint(acc,2*blen); printf("\n");
        }

        //printf("debug: acc="); amsclil1_fb32_liprint(acc,2*blen); printf("\n");

        //bm0 = (bm0**2) % md
        amsclil1_fb32_mult(bm0,bm0,bm1,wrk,blen);
        Kbm = amsclil1_fb32_msi(bm1,2*blen)+1;
        amsclil1_fb32_barrettmod(bm1,mdb,bm0,mu,wrk,Kmd,Kbm,2*blen);

        //amsclil1_fb32_div(bm1,mdb,ra,bm0,wrk,2*blen);

    }


    return;
}


void amsclil1_fb32_ncrandint(uint32_t *a, long blen)
{
    long I;
    uint32_t r;
    for(I=0;I<blen;I++)
    {
        r = rand();
        a[I] = r;
    }
    return;
}